/*
 * mdh (MailDooHicky), a GTK+-based toolbar.
 *
 * Copyright (c) 2003-2005 by Mike Hokenson <mdh at gozer dot org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include <gtk/gtk.h>

#include "mdh.h"

#include "panel.h"

#include "panel_cpu.h"

#include "util.h"

#include "gtop.h"

#if _MDH_HAS_CPU

#define SHOW_BOLD(p, v) \
    (mdh_panel_flags_bold(p) && mdh_panel_cpu_threshold(p) > 0 && \
     v >= mdh_panel_cpu_threshold(p))

typedef struct {
    gdouble loadavg[3];
    guint64 nr_running;
    guint64 nr_tasks;

    guint   ncpu;

    gdouble total,
            total_;
    gdouble user,
            user_;
    gdouble nice,
            nice_;
    gdouble sys,
            sys_;
    gdouble idle,
            idle_;

    gdouble load_r,
            load_a;
} MdhDataCache;

typedef struct {
    MdhDataCache cache;

    gint         threshold;
} MdhPanelData;

static gchar *panel_subst(MdhDataCache *cache, const gchar *fmt);

static gboolean panel_startup(gpointer data, GError **err)
{
    MdhPanel *obj = data;

    g_return_val_if_fail(obj != NULL, FALSE);
    g_return_val_if_fail(obj->display != NULL, FALSE);
    g_return_val_if_fail(obj->interval >= _CPU_MIN, FALSE);
    g_return_val_if_fail(obj->interval <= _CPU_MAX, FALSE);

    if(obj->init)
        return(TRUE);

    debug("panel_startup(): %d", obj->handler);

    obj->init = TRUE;

    return(TRUE);
}

static gboolean panel_shutdown(gpointer data, GError **err)
{
    MdhPanel *obj = data;

    g_return_val_if_fail(obj != NULL, FALSE);

    if(!obj->init)
        return(FALSE);

    debug("panel_shutdown(): %d", obj->handler);

    obj->init = FALSE;

    return(FALSE);
}

static gboolean panel_main(gpointer data, GError **err)
{
    MdhPanel *obj = data;

    MdhDataCache *cache;

    gboolean bold;

    g_return_val_if_fail(obj != NULL, FALSE);

    debug("panel_main(): %d", obj->handler);

    if(!panel_startup(obj, err))
        return(FALSE);

    {
        MdhPanelData *pdata = obj->data;

        cache = &pdata->cache;
    }

    mdh_gtop_get_cpu(cache->loadavg, &cache->nr_running, &cache->nr_tasks,
                     &cache->ncpu,
                     &cache->total, &cache->user, &cache->nice,
                     &cache->sys,   &cache->idle);

    if(cache->user_ && cache->nice_ && cache->sys_) {
        cache->load_r = ((cache->user - cache->user_) +
                         (cache->nice - cache->nice_) +
                         (cache->sys  - cache->sys_)) / obj->interval;

        if(cache->load_r > 100)
            cache->load_r = 100;
    } else
        cache->load_r = 0;

    if((cache->load_a = (cache->loadavg[0] * 100)) > 100)
        cache->load_a = 100;

    bold = SHOW_BOLD(obj, cache->load_a + .5);

    {
        gchar *p = panel_subst(cache, obj->display);
        mdh_panel_set_text_r(obj, bold, "%s", p);
        g_free(p);
    }

    cache->total_ = cache->total;
    cache->user_  = cache->user;
    cache->nice_  = cache->nice;
    cache->sys_   = cache->sys;
    cache->idle_  = cache->idle;

    return(TRUE);
}

#define _real_set(val, new, old)            \
    G_STMT_START {                          \
        if(old) {                           \
            if((val = (new - old)) > 100)   \
                val = 100;                  \
        } else                              \
            val = 0;                        \
    } G_STMT_END

/*
 * %p: load percentage (1 minute average): 43%
 *
 * %1:  1 minute average: 0.43
 * %5:  5 minute average: 0.54
 * %F: 15 minute average: 0.44
 *
 * %t: tasks running (gtop only): 1
 * %T: tasks total (gtop only): 182
 *
 * %c: number of cpus: 1
 *
 * %P: load percentage (real): 18%
 *
 * %o: total (real): 100%
 * %u: user (real): 8%
 * %n: nice (real): 0%
 * %s: sys (real): 10%
 * %i: idle (real): 83%
 */

static gchar *panel_subst(MdhDataCache *cache, const gchar *fmt)
{
    const gchar *p;

    GString *string;

    gdouble total, user, nice, sys, idle;

    g_return_val_if_fail(cache != NULL, NULL);
    g_return_val_if_fail(fmt != NULL, NULL);

    string = g_string_new(NULL);

    _real_set(total, cache->total, cache->total_);
    _real_set(user,  cache->user,  cache->user_);
    _real_set(nice,  cache->nice,  cache->nice_);
    _real_set(sys,   cache->sys,   cache->sys_);
    _real_set(idle,  cache->idle,  cache->idle_);

    for(p = fmt; *p; p++) {
        if(*p == '%' && *(p + 1) && p++) {
            switch(*p) {
                case 'p': /* load percentage (1 minute average): 43% */
                    g_string_append_printf(string, "%3.0f%%", cache->load_a);
                    break;

                case '1': /*  1 minute average: 0.43 */
                    g_string_append_printf(string, "%0.2f", cache->loadavg[0]);
                    break;
                case '5': /*  5 minute average: 0.54 */
                    g_string_append_printf(string, "%0.2f", cache->loadavg[1]);
                    break;
                case 'F': /* 15 minute average: 0.44 */
                    g_string_append_printf(string, "%0.2f", cache->loadavg[2]);
                    break;

                case 't': /* tasks running (gtop only): 1 */
                    g_string_append_printf(string, "%llu", cache->nr_running);
                    break;
                case 'T': /* tasks total (gtop only): 182 */
                    g_string_append_printf(string, "%llu", cache->nr_tasks);
                    break;

                case 'c': /* number of cpus: 1 */
                    g_string_append_printf(string, "%u", cache->ncpu);
                    break;

                case 'P': /* load percentage (real): 27% */
                    g_string_append_printf(string, "%3.0f%%", cache->load_r);
                    break;

                case 'o': /* total (real): 100% */
                    g_string_append_printf(string, "%3.0f%%", total);
                    break;
                case 'u': /* user (real): 8% */
                    g_string_append_printf(string, "%3.0f%%", user);
                    break;
                case 'n': /* nice (real): 0% */
                    g_string_append_printf(string, "%3.0f%%", nice);
                    break;
                case 's': /* sys (real): 10% */
                    g_string_append_printf(string, "%3.0f%%", sys);
                    break;
                case 'i': /* idle (real): 83% */
                    g_string_append_printf(string, "%3.0f%%", idle);
                    break;

                default:
                    g_string_append_c(string, *p);
                    break;
            }
        } else
            g_string_append_c(string, *p);
    }

    return(g_string_free(string, FALSE));
}

gint mdh_panel_cpu_threshold(MdhPanel *obj)
{
    g_return_val_if_fail(obj != NULL, 0);
    g_return_val_if_fail(obj->data != NULL, 0);

    return(((MdhPanelData *)obj->data)->threshold);
}

void mdh_panel_cpu_threshold_set(MdhPanel *obj, gint threshold)
{
    g_return_if_fail(obj != NULL);
    g_return_if_fail(obj->data != NULL);

    ((MdhPanelData *)obj->data)->threshold = threshold;
}

static MdhPanelFuncs funcs = {
    panel_main,
    panel_startup,
    panel_shutdown,
    NULL,
    NULL
};

MdhPanel *mdh_panel_cpu_new(const gchar *display,
                            const gchar *command,
                            gboolean enabled)
{
    MdhPanel *obj = mdh_panel_new();

    g_return_val_if_fail(obj != NULL, NULL);

    obj->display  = (display) ? g_strdup(display) : g_strdup(_CPU_DIS);
    obj->command  = (command) ? g_strdup(command) :
                                g_strdup_printf("%s -e top", mdh_get_xterm());
    obj->tooltip  = g_strdup("Cpu");
    obj->interval = _CPU_DEF;
    obj->enabled  = enabled;
    obj->funcs    = funcs;

    obj->data     = g_new0(MdhPanelData, 1);

    ((MdhPanelData *) obj->data)->threshold = 0;

    return(obj);
}

#endif /* _MDH_HAS_CPU */
