/*
 * mdh (MailDooHicky), a GTK+-based toolbar.
 *
 * Copyright (c) 2003-2005 by Mike Hokenson <mdh at gozer dot org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include <gtk/gtk.h>

#include "mdh.h"

#include "panel.h"

#include "panel_mem.h"

#include "util.h"

#include "gtop.h"

#if _MDH_HAS_MEM

#define SHOW_BOLD(p, v) \
    (mdh_panel_flags_bold(p) && mdh_panel_mem_threshold(p) > 0 && \
     v >= mdh_panel_mem_threshold(p))

typedef struct {
    guint64 total,
            used,
            free;

    gdouble free_p;

    guint64 real_total,
            real_used,
            real_free;

    gdouble real_free_p;

    guint64 swap_total,
            swap_used,
            swap_free;

    gdouble swap_free_p;
} MdhDataCache;

typedef struct {
    gboolean buffers;

    gint     threshold;
} MdhPanelData;

static gchar *panel_subst(MdhDataCache *cache, const gchar *fmt);

static gboolean panel_startup(gpointer data, GError **err)
{
    MdhPanel *obj = data;

    g_return_val_if_fail(obj != NULL, FALSE);
    g_return_val_if_fail(obj->display != NULL, FALSE);
    g_return_val_if_fail(obj->interval >= _MEM_MIN, FALSE);
    g_return_val_if_fail(obj->interval <= _MEM_MAX, FALSE);

    if(obj->init)
        return(TRUE);

    debug("panel_startup(): %d", obj->handler);

    obj->init = TRUE;

    return(TRUE);
}

static gboolean panel_shutdown(gpointer data, GError **err)
{
    MdhPanel *obj = data;

    g_return_val_if_fail(obj != NULL, FALSE);

    if(!obj->init)
        return(FALSE);

    debug("panel_shutdown(): %d", obj->handler);

    obj->init = FALSE;

    return(FALSE);
}

static gboolean panel_main(gpointer data, GError **err)
{
    MdhPanel *obj = data;

    MdhDataCache cache;

    guint64 buffers;

    gboolean bold;

    g_return_val_if_fail(obj != NULL, FALSE);

    debug("panel_main(): %d", obj->handler);

    if(!panel_startup(obj, err))
        return(FALSE);

    mdh_gtop_get_mem(&cache.real_total,
                     &cache.real_used,
                     &cache.real_free,
                     &buffers);

    mdh_gtop_get_swap(&cache.swap_total,
                      &cache.swap_used,
                      &cache.swap_free);

    if(!mdh_panel_mem_buffers(obj))
        cache.real_used -= buffers;

    cache.total = cache.real_total + cache.swap_total;
    cache.used  = cache.real_used  + cache.swap_used;
    cache.free  = cache.real_free  + cache.swap_free;

#define _P(u, t) (100 - (((gdouble) u / t * 100) + .5))

    cache.free_p      = _P(cache.used,      cache.total);
    cache.real_free_p = _P(cache.real_used, cache.real_total);
    cache.swap_free_p = _P(cache.swap_used, cache.swap_total);

    bold = SHOW_BOLD(obj, (100 - cache.free_p) + .5);

    {
        gchar *p = panel_subst(&cache, obj->display);
        mdh_panel_set_text_r(obj, bold, "%s", p);
        g_free(p);
    }

    return(TRUE);
}

#define DVAL(v) ((gdouble) (v) / 1048576)

/*
 * %u: used: 431MB
 * %U: used percent: 22%
 * %f: free: 1207MB
 * %F: free percent: 78%
 * %t: total: 1991MB
 *
 * %r: real used: 429MB
 * %R: real used percent: 43%
 * %e: real free: 229MB
 * %E: real free percent: 57%
 * %T: real total: 1010MB
 *
 * %s: swap used: 2MB
 * %S: swap used percent: 1%
 * %w: swap free: 978MB
 * %W: swap free percent: 99%
 * %O: swap total: 980MB
 */

static gchar *panel_subst(MdhDataCache *cache, const gchar *fmt)
{
    const gchar *p;

    GString *string;

    g_return_val_if_fail(cache != NULL, NULL);
    g_return_val_if_fail(fmt != NULL, NULL);

    string = g_string_new(NULL);

    for(p = fmt; *p; p++) {
        if(*p == '%' && *(p + 1) && p++) {
            switch(*p) {
                case 'u': /* used: 431MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->used));
                    break;
                case 'U': /* used percent: 22% */
                    g_string_append_printf(string, "%01.0f%%",
                                           100 - cache->free_p);
                    break;

                case 'f': /* free: 1207MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->free));
                    break;
                case 'F': /* free percent: 78% */
                    g_string_append_printf(string, "%01.0f%%",
                                           cache->free_p);
                    break;

                case 't': /* total: 1991MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->total));
                    break;

                case 'r': /* real used: 429MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->real_used));
                    break;
                case 'R': /* real used percent: 43% */
                    g_string_append_printf(string, "%01.0f%%",
                                           100 - cache->real_free_p);
                    break;

                case 'e': /* real free: 229MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->real_free));
                    break;
                case 'E': /* real free percent: 57% */
                    g_string_append_printf(string, "%01.0f%%",
                                           cache->real_free_p);
                    break;

                case 'T': /* real total: 1010MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->real_total));
                    break;

                case 's': /* swap used: 2MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->swap_used));
                    break;
                case 'S': /* swap used percent: 1% */
                    g_string_append_printf(string, "%01.0f%%",
                                           100 - cache->swap_free_p);
                    break;

                case 'w': /* swap free: 978MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->swap_free));
                    break;
                case 'W': /* swap free percent: 99% */
                    g_string_append_printf(string, "%01.0f%%",
                                           cache->swap_free_p);
                    break;

                case 'O': /* swap total: 980MB */
                    g_string_append_printf(string, "%1.0fMB",
                                           DVAL(cache->swap_total));
                    break;

                default:
                    g_string_append_c(string, *p);
                    break;
            }
        } else
            g_string_append_c(string, *p);
    }

    return(g_string_free(string, FALSE));
}

gboolean mdh_panel_mem_buffers(MdhPanel *obj)
{
    g_return_val_if_fail(obj != NULL, 0);
    g_return_val_if_fail(obj->data != NULL, 0);

    return(((MdhPanelData *)obj->data)->buffers);
}

void mdh_panel_mem_buffers_set(MdhPanel *obj, gboolean buffers)
{
    g_return_if_fail(obj != NULL);
    g_return_if_fail(obj->data != NULL);

    ((MdhPanelData *)obj->data)->buffers = buffers;
}

gint mdh_panel_mem_threshold(MdhPanel *obj)
{
    g_return_val_if_fail(obj != NULL, 0);
    g_return_val_if_fail(obj->data != NULL, 0);

    return(((MdhPanelData *)obj->data)->threshold);
}

void mdh_panel_mem_threshold_set(MdhPanel *obj, gint threshold)
{
    g_return_if_fail(obj != NULL);
    g_return_if_fail(obj->data != NULL);

    ((MdhPanelData *)obj->data)->threshold = threshold;
}

static MdhPanelFuncs funcs = {
    panel_main,
    panel_startup,
    panel_shutdown,
    NULL,
    NULL
};

MdhPanel *mdh_panel_mem_new(const gchar *display,
                            const gchar *command,
                            gboolean enabled)
{
    MdhPanel *obj = mdh_panel_new();

    g_return_val_if_fail(obj != NULL, NULL);

    obj->display  = (display) ? g_strdup(display) : g_strdup(_MEM_DIS);
    obj->command  = (command) ? g_strdup(command) : g_strdup("");
    obj->tooltip  = g_strdup("Mem");
    obj->interval = _MEM_DEF;
    obj->enabled  = enabled;
    obj->funcs    = funcs;

    obj->data     = g_new(MdhPanelData, 1);

    ((MdhPanelData *) obj->data)->buffers   = TRUE;
    ((MdhPanelData *) obj->data)->threshold = 0;

    return(obj);
}

#endif /* _MDH_HAS_MEM */
